using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Reflection;

namespace SimpleDoubleDispatchSample {
    public class Program {
        private static void Main(string[] args) {
            Collection<AbstractVisitor> visitors = new Collection<AbstractVisitor>();
            Collection<Pet> pets = new Collection<Pet>();

            pets.Add(new Fish());
            pets.Add(new Dog());

            visitors.Add(Feeder.Instance);
            visitors.Add(Walker.Instance);

            // Visit each of the pets.
            foreach (Pet pet in pets) {
                foreach (AbstractVisitor visitor in visitors) {
                    visitor.Visit(pet);
                }
            }

            // Check the results.
            foreach(Pet pet in pets) {
                Console.Out.WriteLine(pet.GetType().Name);
                foreach(String note in pet.Visitors) {
                    Console.Out.WriteLine("\t{0}", note);
                }
            }
        }
    }

    /// <summary>
    /// Concrete visitor, a pet feeder.
    /// </summary>
    public class Feeder : AbstractVisitor {
        private static readonly Feeder instance = new Feeder();
        
        protected Feeder() : base() {}

        public static Feeder Instance {
            get { return instance; }
        }


        public void Visit(Dog dog) {
            // Feed the dog
            dog.Visitors.Add("Fed the dog");
        }

        public void Visit(Fish fish) {
            // Feed the fish
            fish.Visitors.Add("Fed the fish");
        }
    }

    /// <summary>
    /// Concrete visitor, a pet walker.
    /// </summary>
    public class Walker : AbstractVisitor {
        private static readonly Walker instance = new Walker();

        protected Walker() : base() { }

        public static Walker Instance {
            get { return instance; }
        }

        public void Visit(Dog dog) {
            dog.Visitors.Add("Walked the dog");
        }

        // Override to handle all unknown/unsupported types.
        public override void VisitGeneric(object pet) {
            Console.Out.WriteLine("Can't walk {0}!", pet.GetType().Name);
        }
    }

    /// <summary>
    /// Base class for pets.
    /// </summary>
    public abstract class Pet {
        private readonly Collection<string> visitors;

        public Collection<string> Visitors {
            get { return visitors; }
        }

        protected Pet() {
            visitors = new Collection<string>();
        }
    }

    /// <summary>
    /// A pet fish.
    /// </summary>
    public class Fish : Pet {}

    /// <summary>
    /// A pet doc.
    /// </summary>
    public class Dog : Pet {}

    /// <summary>
    /// Abstract base class for visitors.
    /// </summary>
    public abstract class AbstractVisitor {
        private readonly Dictionary<long, MethodInfo> dispatch;

        protected AbstractVisitor() {
            dispatch = new Dictionary<long, MethodInfo>();
            MethodInfo[] methods = GetType().GetMethods();

            foreach(MethodInfo method in methods) {
                if(method.Name == "Visit") {
                    ParameterInfo[] parameters = method.GetParameters();

                    if(parameters.Length == 1) {
                        long key = (parameters[0].ParameterType.GetHashCode());
                        
                        dispatch.Add(key, method);
                    }
                }
            }
        }

        public virtual void Visit(object pet) {
            Int64 hash = pet.GetType().GetHashCode();
            if (dispatch.ContainsKey(hash)) {
                dispatch[hash].Invoke(this, new object[] { pet });                
            }
            else {
                VisitGeneric(pet);
            }
        }

        public virtual void VisitGeneric(object pet) {
            Console.Out.WriteLine("Generic visit for {0}", pet.GetType().Name);
        }
    }
}